(**************************************************************************)
(*                                                                        *)
(*      whitelister : a Whitelister Policy Daemon for Postfix             *)
(*      ~~~~~~~~~~~                                                       *)
(*                                                                        *)
(*  Copyright (C) 2005 Polytechnique.org                                  *)
(*  Author : Pierre Habouzit <pierre.habouzit@m4x.org>                    *)
(*  ____________________________________________________________________  *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation; either version 2 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program; if not, write to the Free Software           *)
(*  Foundation, Inc.,                                                     *)
(*  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA               *)
(*                                                                        *)
(**************************************************************************)

(* types *)

type config = {
  mutable pidf:  string;
  mutable sock:  Unix.sockaddr;
  mutable rbl:   string list;
  mutable rhbl:  string list;
  mutable user:  string;
  mutable group: string;
}

(* Checker *)

let log s =
  let lg = Syslog.openlog ~facility:`LOG_MAIL ~flags:[`LOG_PID] "whitelister" in
  Syslog.syslog lg `LOG_INFO s;
  Syslog.closelog lg

let checker cfg ic oc =
  let pcy = Policy.read ic in
  let ok = (
    try
      Rules.check_rbl cfg.rbl pcy;
      Rules.check_rhbl cfg.rhbl pcy;
      log "Clean";
      true
    with
      | Rules.Dirty s -> log ("Dirty (" ^ s ^ ")"); false
  ) in
  Printf.fprintf oc "action=%s\n\n" (if ok then "OK" else "DUNNO");
  flush oc

(* Configuration *)

let empty_config () = {
  pidf  = "/var/run/whitelister.pid" ;
  sock  = Unix.ADDR_INET (Unix.inet_addr_of_string "127.0.0.1", 10000) ;
  rbl   = [] ;
  rhbl  = [] ;
  user  = "nobody" ;
  group = "nogroup" ;
}
    
let update_config cfg file =
  let ic  = open_in file in
  try
    while true do
      let line = input_line ic in
      if String.length line > 0 && line.[0] <> '#' then
        match Str.split (Str.regexp "[ \t:]+") line with
          | [] -> ()
          | ["pidf"; f]     -> cfg.pidf  <- f
          | ["sock"; s]     -> cfg.sock  <- Unix.ADDR_UNIX s
          | ["sock"; ip; p] -> cfg.sock  <- Unix.ADDR_INET (Unix.inet_addr_of_string ip, int_of_string p)
          | ["rbl";  h]     -> cfg.rbl   <- h::cfg.rbl
          | ["rhbl"; h]     -> cfg.rhbl  <- h::cfg.rhbl
          | ["user"; u]     -> cfg.user  <- u
          | ["group"; g]    -> cfg.group <- g
          | _ -> prerr_string (Printf.sprintf "`%s' is not a valid config line\n" line)
    done;
    assert false
  with End_of_file -> cfg

let read_config () =
  let found_files = List.filter Sys.file_exists ["/etc/whitelister.conf" ; "whitelister.conf"] in
  List.fold_left update_config (empty_config ()) found_files

(* Server thingies *)

open Unix

let write_pid file =
  try
    let oc = open_out file in
    output_string oc (Printf.sprintf "%i\n" (getpid ()));
    close_out oc
  with Sys_error _ ->
        prerr_endline ("Cannot write my pid in the pidfile "^file);
        exit 1
    
let drop_privs cfg =
  if geteuid () = 0 then (
    setgid (getgrnam cfg.group).gr_gid;
    setuid (getpwnam cfg.user).pw_uid
  )
              
let cleanup_sock = function
  | Unix.ADDR_INET _ -> ()
  | Unix.ADDR_UNIX u -> (try Unix.unlink u with _ -> ())

(* Main LOOP *)

let _ =
  let cfg = read_config () in
  write_pid cfg.pidf;
  match fork () with
    | 0 ->
        write_pid cfg.pidf;
        drop_privs cfg; 
        cleanup_sock cfg.sock;
        Unix.establish_server (checker cfg) cfg.sock
    | _ -> exit 0

